#
#!/bin/bash
#
# functions_vk.bash
#
# Volker's function library for bash shell scripts.
# This file must be sourced, not executed!!
#
# Load it at the start of the script with
#    checkbash2=1 requireversion=22 source functions_vk.bash
# The two variable assignments may be left out if that functionality is not
# required.
#
# Copyright (C) by Volker Kuhlmann <v.kuhlmann@elec.canterbury.ac.nz>
# Released under the terms of the GNU General Public License (GPL) Version 2.
# See http://www.gnu.org/ for details.
#
# Volker Kuhlmann
#   16 Oct 1999
#   28, 29 Mar; 12 Apr; 23, 27 May; 27 Jul; 21 Aug; 10, 24, 25 Oct 2000
#   2, 4, 6, 19 Sep 2001
#   26 Jan; 5 Apr; 16 May; 1 Sep; 9 Nov; 5 Dec 2002
#   6 Mar 2003
#

VERSION_FUNCLIB="VK 1.7.5, 05 Mar 2002"
VERSION_FUNCLIB_NUM="28"   # this integer will be incremented with each change


#### Check bash is version 2 ####
#
# Only if checkbash2 is non-empty: check bash version, and run under 2.x
# if the current one is 1.x and a 2.x can be found. Error if no 2.x available.
#
if [ -n "$checkbash2" ]; then
    if [ ${BASH_VERSION%%.*} -lt 2 ]; then
	b="`type -p -a bash2 bash-2 bash_2 bash-2.x bash-2.X bash-2.03.0 | head -1`"
	if [ -x "$b" ]; then
	    v="`$b -version </dev/null | grep 'ersion 2'`"
	    if [ -z "$v" ]; then
		echo "$b is not a bash version 2.x"
		exit 49
	    fi
    	    exec $b -c "`cat $0`" "$0" "$@"
	else
	    exec 1>&2
    echo "This script requires bash version 2. Sorry, but"
    echo "  - bash 1 is too pathetic"
    echo "  - ksh is neither syntax compatible nor freely available"
    echo "  - pdksh is no advantage over bash 2"
    echo "  - tcsh lacks essential features (although some things are nicer"
    echo "    than in bash)"
	    exit 49
	fi
    fi
fi

#### Check library version number ####
#
if [ -n "$requireversion" ]; then
    : ${lib:=functions_vk.bash}
    : ${url:=http://volker.dnsalias.net/soft/script/functions_vk.bash}
    if [ "${VERSION_FUNCLIB_NUM:-0}" -lt "$requireversion" ]; then
	echo 1>&2 "\
Your version of $lib ($VERSION_FUNCLIB)
in $(dirname $(type -p $lib))
is too old for ${0##*/}. Please get a new one from e.g.
$url"
	exit 51
    fi
    unset lib url
fi


#### Error/Exit codes
#
# User progam needs to redefine exitwith (by copying this exitwith() and
# extending it) and provide error messages for new error codes introduced by
# the user program. The user program may use the codes defined in
# exitwith_funclib() as well.
#
exitwith_funclib() {
  case "$1" in
    ErrUsage|ErrHelp)
	# usage already displayed
	exit 1;;
    ErrNoExist)
    	echo "'$2' does not exist."
	exit 2;;
    ErrInput)
    	echo "Does not exist or can't read: '$2'."
	exit 2;;
    ErrNoDir)
    	echo "Directory '$2' does not exist (or is not readable)."
	exit 2;;
    ErrNoFile)
    	echo "File '$2' does not exist (or is not readable)."
    	#echo "'$2' is not a file or doesn't exist."
	exit 2;;
 # more codes in here  --->
 # <--- more codes in here
    ErrBadoption)
    	echo "Bad option '$2'."
	echo "Call with -h for help."
	exit 9;;
    ErrMissingParameter)
	echo "A required parameter for option $2 is missing."
	echo "Call with -h for help."
	exit 9;;
  esac
}
exitwith() {
    exec 1>&2  # write stdout on stderr instead
    exitwith_funclib "$@"
  case "$1" in
 # more codes in here  --->
 # <--- more codes in here
    *)
	echo "Error: exitwith() was called with unknown error code '$1',"
	echo "       or user application did not define exitwith()."
	exit 19;;
  esac
}


#### Command arg scanning ####
#
checkargmm() {
    test "$1" != "--" -a "$2" != "--" \
      && test \( "${1:0:1}" != "-" -o "$1" == "-" \) \
              -a \( "${2:0:1}" != "-" -o "$2" == "-" \)
    # uncomment after && above to allow arguments starting with "-" to options
    # (a single "-" is always allowed)
    #true  # uncomment this to allow "--" as argument to an option
}
checkarg2() { # 1 required
    test $# -ge 2 && checkargmm "$2" && return
    exitwith ErrMissingParameter "$1"
}
checkarg3() { # 2 required
    test $# -ge 3 && checkargmm "$2" "$3" && return
    exitwith ErrMissingParameter "$1"
}
checkarg2o() { # 0 required, 1 optional (returning true if present)
    test $# -ge 2 || return  # no optional arg if $2 missing
    test "$2" == "${2#-}"    # no optional arg if $2 starts with "-"
}
checkarg3o() { # 1 required, 1 optional (returning true if present)
    checkarg2 "$@"
    test $# -ge 3 || return  # no optional arg if $3 missing
    test "$3" == "${3#-}"    # $3 is no optional arg if starts with "-"
}


#### GNU programs ####
#
# Make sure we have GNU versions of some programs PROG available as gnuPROG.
# If the GNU version is not available gnuPROG is undefined.
# Use: if you need a GNU version in your script, use gnuPROG. On a non-GNU unix
# system install the GNU programs as gnuPROG.
#
# Bash aliases are somewhat crappy. In non-posix mode (#!/bin/bash, or
# set +p posix) a command defined as alias will not be found. Even in posix
# mode, the alias must be defined before any function is defined in which it is
# used, this means find_GNU_prog() must be called before any functions are
# defined. Trying functions instead.
#
find_GNU_prog1() {
    # THIS HAS to be changed somehow. Aliases are utter *crap* in bash.
    local p
    type -p gnutar >/dev/null || {
      p="`type -p tar`" && "$p" --version >/dev/null 2>&1 && alias gnutar="$p"
    }
    type -p gnucpio >/dev/null || {
      p="`type -p cpio`" && "$p" --version >/dev/null 2>&1 && alias gnucpio="$p"
    }
    type -p gnuawk >/dev/null || {
      p="`type -p gawk`" && "$p" --version >/dev/null 2>&1 \
	    && alias gnuawk="$p" \
      || p="`type -p awk`" \
	    && test "`$p --version </dev/null >/dev/null`" != "" \
	    && alias gnuawk="$p"
    }
    # doesn't recognise /bin/ls as GNU (type -p ls returns nothing)
    type -p gnuls >/dev/null || {
      p="`type -p ls`" && "$p" --version >/dev/null 2>&1 && alias gnuls="$p"
    }
    type -p gnufind >/dev/null || {
      p="`type -p find`" && "$p" --version >/dev/null 2>&1 && alias gnufind="$p"
    }
    type -p gnuxargs >/dev/null || {
      p="`type -p xargs`" && "$p" --version >/dev/null 2>&1 \
	    && alias gnuxargs="$p"
    }
    type -p gnutouch >/dev/null || {
      p="`type -p touch`" && "$p" --version >/dev/null 2>&1 \
	    && alias gnutouch="$p"
    }
    type -p gnudiff >/dev/null || {
      p="`type -p diff`" && "$p" --version >/dev/null 2>&1 && alias gnudiff="$p"
    }
}
find_GNU_prog_() {
    local p prog="$1"; shift
    type -p "$prog" >/dev/null || {
	while [ $# -gt 0 ]; do
	    p="`type -p "$1"`" \
	    && "$p" --version </dev/null >/dev/null 2>&1 \
	    && {
	    	eval function "$prog" '{ "'$p'" "$@"; }'
		return 1
	    }
	    shift
	done
    }
}
find_GNU_prog() {
    local p
    find_GNU_prog_ gnutar gtar tar
    find_GNU_prog_ gnucpio gcpio cpio
    #find_GNU_prog_ gnuawk gawk awk  #test fails for solaris awk
    find_GNU_prog_ gnuawk gawk || {
    	p="`type -p awk`"
	test "`$p --version </dev/null >/dev/null`" != "" \
	    && eval function gnuawk '{ "'$p'" "$@"; }'
    }
    # doesn't recognise /bin/ls as GNU (type -p ls returns nothing)
    find_GNU_prog_ gnuls ls /bin/ls
    find_GNU_prog_ gnufind gfind find
    find_GNU_prog_ gnuxargs gxargs xargs
    find_GNU_prog_ gnutouch gtouch touch
    find_GNU_prog_ gnudiff gdiff diff
}
# Find a better version of awk
find_better_prog() {
    local p
    type -p AWK >/dev/null || {
    	#p="`type -p gnuawk gawk nawk | head -1`" && alias AWK="$p"
    	p="`type -p gnuawk gawk nawk | head -1`" && \
	    eval 'AWK() { '"$p"' "$@"; }'
    }
}
debug_find_GNU_prog() {
    echo "gnutar=  `type gnutar 2>/dev/null`"
    echo "gnucpio= `type gnucpio 2>/dev/null`"
    echo "gnuawk=  `type gnuawk 2>/dev/null`"
    echo "gnuls=   `type gnuls 2>/dev/null`"
    echo "gnufind= `type gnufind 2>/dev/null`"
    echo "gnuxargs=`type gnuxargs 2>/dev/null`"
    echo "gnutouch=`type gnutouch 2>/dev/null`"
    echo "gnudiff= `type gnudiff 2>/dev/null`"
    echo "AWK=     `type AWK 2>/dev/null`"
}


#### Logging ####
#
# Aid for screen logging of executed commands (similar to "set echo" in tcsh?)
    #set -x; "$@"  # affects all following
    #(set -x; "$@")  # works but starts another shell, and logs 2 lines if
    	    	    #	command is a shell function
    #set -x; "$@"; set +x  # logs "set +x"
# Log + run given command
Log() {
    #echo "$@"; "$@"
    local _e=
    test "$1" = "-e" && { shift; _e=eval; }
    echo "$cmdlogprefix""$@"
    $_e "$@"
}
# Log given command, and run it unless dry-run is on
DryLog() {
    local _e=
    test "$1" = "-e" && { shift; _e=eval; }
    echo "$cmdlogprefix""$@"
    ifset dryrun && return
    $_e "$@"
}
# Either log (if dry-run is on) or run given command
RunLog() {
    if ifset dryrun; then
    	echo "$cmdlogprefix""$@"
    else
	local _e=
	test "$1" = "-e" && { shift; _e=eval; }
	$_e "$@"
    fi
}
# until version 26:
#Log_e() { Log -e "$@"; }
#Nolog() { RunLog "$@"; }


#
# show $0 and all cmd line arguments
show_args() {
    local n i=0
    echo "line (#=$#): ""$@"
    while [ $i -le $# ]; do
    	echo "$i: '${!i}'"
	i=$(($i + 1))
    done
}
debug_show_args() {
    echo --------; show_args "$@"; echo --------
    show_args "${@:3}"; echo --------
    bla=("${@:3}"); show_args x "${bla[@]}"; echo --------
}


#### Boolean flags ####
#
# return true if the variable is set; "set" means != "" but not "0"
ifset() {
    if [ "$1" == "!" ]; then
	test ! \( -n "${!2}" -a ! 0${!2} -eq 0 \)
    else
	test -n "${!1}" -a ! 0${!1} -eq 0
    fi
}
ifnset() {
    ifset ! $1
}
debug_ifset() {
    local v1=0 v2=1 v3="" v4; unset v4
    ifset v1 && echo v1; ifset v2 && echo v2; ifset v3 && echo v3
    ifset v4 && echo v4
    echo -
    ifset ! v1 && echo v1; ifset ! v2 && echo v2; ifset ! v3 && echo v3
    ifset ! v4 && echo v4
    echo -
    ifnset v1 && echo v1; ifnset v2 && echo v2; ifnset v3 && echo v3
    ifnset v4 && echo v4
}


#### Files' status ####

# file is ((plain or symlink) and readable)
test_readablefile() {
    test \( -f "$1" -o -L "$1" \) -a -r "$1"
}

checkforexist() {
    test -e $1 || exitwith ErrNoExist "$1"
}

checkfordir() {
    test -d "$1" -a -r "$1" || exitwith ErrNoDir "$1"
}

checkforfile() {
    test -f "$1" -a -r "$1" || exitwith ErrNoFile "$1"
}


#### Read user resources ####

# Read user resources $1; if $2 = -home also try to read $HOME/$1
# Return msg in $rcmsg
readresources() {
    # Include other resources, can be called form resource file to nest
    include_resource() {
	test_readablefile "$1" && source "$1"
    }

    if test_readablefile "$1"; then
	source "$1"
	rcmsg="Resources read from ${1/#$HOME/~}"
    elif test "$2" = "-home" && test_readablefile "$HOME/$1"; then
	source "$HOME/$1"
	rcmsg="Resources read from ~/$1"
    else
	rcmsg="Resources not found (${1/#$HOME/~})"
    fi
    #echo $rcmsg
}


#### Creating directories and symlinks, etc ####

_showcreate=1
_drycreate=0
makedir() {
    while [ $# -gt 0 ]; do
	if [ ! -d "$1" ]; then
    	    ifset _showcreate || ifset _drycreate && echo "mkdir: $1"
	    ifnset _drycreate && mkdir -p "$1"
	fi
	shift
    done
}
makelink() {
    if [ ! -L "$1" ]; then
    	ifset _showcreate || ifset _drycreate && echo "link:  $1 -> $2"
	ifnset _drycreate && ln -s "$2" "$1"
    fi
}
makelinkdel() { # as makelink, but deletes link first if it exists
    ifset _showcreate || ifset _drycreate && echo "link:  $1 -> $2"
    ifnset _drycreate && {
    	if [ -L "$1" ]; then rm "$1"; fi
    	ln -s "$2" "$1"
    }
}
deldir() {
    while [ $# -gt 0 ]; do
	if [ -d "$1" ]; then
    	    ifset _showcreate || ifset _drycreate && echo "rmdir: $1"
	    ifnset _drycreate && rmdir "$1"
	fi
	shift
    done
}
makemove() {
    ifset _showcreate || ifset _drycreate && echo "move:  ""$@"
    ifnset _drycreate && mv "$@"
    # mv does not preserve permissions across filesystems!
}


#### Paths ####

# replace $HOME-part of path with ~
hometilde() {
    echo "${1/#$HOME/~}"
}

# strip trailing "/" from path
slashpath() {
    echo "${1%/}"
}

# path components
# tcsh -fc 'set f=path/base.ext; echo $f:r $f:h $f:t $f:r:t $f:e'
#   => path/base path base.ext base ext
# Fixed bug with path_h() (was 1%%/*), 5Apr02.
path_r() { echo "${1%.*}"; }	# tcsh: root
path_h() { echo "${1%/*}"; }	# tcsh: head
path_t() { echo "${1##*/}"; }	# tcsh: tail
path_b() { path_r "${1##*/}"; } # tcsh: tail root
path_e() { echo "${1##*.}"; }	# tcsh: ext
path_d() { # directory part (similar to dirname)
    if [ "${1%/*}" == "$1" ]; then
    	echo "."
    else
    	echo "${1%/*}"
    fi
}
path_debug() {
    echo "pa/h/base.ext  root  `path_r pa/h/base.ext`"
    echo "pa/h/base.ext  head  `path_h pa/h/base.ext`"
    echo "pa/h/          head  `path_h pa/h/`"
    echo "pa/h/base.ext  tail  `path_t pa/h/base.ext`"
    echo "pa/h/base.ext  r, t  `path_b pa/h/base.ext`"
    echo "pa/h/base.ext  ext   `path_e pa/h/base.ext`"
    echo "pa/h/file      dir   `path_d pa/h/file`"
    echo "pa/h/          dir   `path_d pa/h/`"
    echo "path/          dir   `path_d path/`"
    echo "path           dir   `path_d path`"
    echo "file           dir   `path_d file`"
}


### Simple stack for shell options ###

# Usage: push_shopt nocaseglob nullglob [..]
#        pop_shopt nocaseglob nullglob [..]

shopt_stack=""
push_shopt() {
    while [ $# -gt 0 ]; do
    	v='u'
	shopt -q "$1" && v='s'
	shopt_stack="$shopt_stack$v"
	shift
    	#echo "push: $shopt_stack"
    done
}
pop_shopt() {
    #echo "pop in: $shopt_stack; " "$@"
    if [ $# -gt ${#shopt_stack} ]; then
	echo 1>&2 "pop_shopt(): stack underflow with '${@:${#shopt_stack}+1}'"
	shopt_stack=""; false; return
    fi
    if [ $# -gt 1 ]; then
	pop_shopt "${@:2}"
    fi
    #echo "popping: $1"
    shopt -${shopt_stack:${#shopt_stack}-1} "$1"
    shopt_stack="${shopt_stack:0:${#shopt_stack}-1}"
    #echo "pop : $shopt_stack"
}
debug_shopt_stack() {
    shopt -p nocaseglob nullglob promptvars shift_verbose sourcepath extglob
    push_shopt nocaseglob nullglob promptvars shift_verbose sourcepath
    shopt -s nocaseglob nullglob
    shopt -u sourcepath
    shopt -p nocaseglob nullglob promptvars shift_verbose sourcepath extglob
    s="$shopt_stack"
    pop_shopt nocaseglob nullglob promptvars shift_verbose sourcepath extglob
    shopt_stack="$s"
    pop_shopt nocaseglob nullglob promptvars shift_verbose sourcepath
    shopt -p nocaseglob nullglob promptvars shift_verbose sourcepath extglob
}


#### CDROM ####

# Print the number of CD blocks (in 2048 bytes) of the given cdrom device or
# iso file to stdout.
# iso_numblocks DEVICE_OR_ISOFILE [FUNCTION_TO_CALL]
# iso_numblocks set FUNCTION_TO_CALL
iso_numblocks() {
    if [ "$1" == "set" ]; then
    	# set method by which to obtain block count
    	case "$2" in
	    isoinfo)	_func_iso_getblocks="isoinfo_blocks";;
	    *)	    	_func_iso_getblocks="$2";;
    	esac
	return
    fi
    ${2-$_func_iso_getblocks} "$1"
}
iso_numblocks set isoinfo

# *) var=`exit 1`  exits the subshell with code 1, not the parent
# *) local var=`return 1` doesn't set $?; but local var
# *) var=`return 1` sets $?

# Using program isoinfo (shipped with cdrecord) to obtain block count.
isoinfo_blocks() {
    local s
    s=`isoinfo -i "$1" -d`
    # check return code of isoinfo
    test $? -ne 0 && exitwith ErrInput "$1"
    s=`echo "$s" | awk '
    	$0 ~ /.ogical block size/ {if ($NF != 2048) exit 1}
	$0 ~ /.olume size/ {print $NF}
    	'`
    echo $s
}

# Return all blocks of the given cdrom device or iso file to stdout.
# Args: DEVICE_OR_ISOFILE  (can't be a pipe!)
# If $1 is a plain file, we could run cat instead - downside of that is the we
# return the whole file, not the size of the iso9660 of it (the file may be
# longer or shorter).
iso_blockread() {
    #iso_read_blocks "$1" "`iso_numblocks "$1"`"
    local n ret
    n="`iso_numblocks "$1"`"
    ret=$?
    test $ret -ne 0 && exit $ret
    iso_read_blocks "$1" "$n"
}

# Read given cdrom device or iso file with given blocksize, saving to given path
# prefix with block number appended. Useful with read errors on parts of the CD.
# Args: DEVICE_OR_ISOFILE BLOCKS_AT_A_TIME START_BLOCK END_BLOCK PATHPREFIX
# (START_BLOCK and PATHPREFIX may be empty.)
iso_blockread_numblocks() {
    local iso="$1" chunk=$2 block=${3:-0} size=$4 prefix="${5:-cdblocks}"
    test -z "$size" &&  size="`iso_numblocks "$iso"`"
    test $chunk = 0 && return
    while [ $block -lt $size ]; do
    	test $(($block + $chunk)) -gt $size && chunk=$(($size - $block))
    	block0="`iso_printf06i $block`"
    	dd if="$iso" bs=2k count="$chunk" skip="$block" of="$prefix$block0" \
	    >/dev/null 2>&1 \
	    || echo "Error: blocks starting $block (chunk=$chunk)" 1>&2
	block=$(($block + $chunk))
    done
}
iso_printf06i() {
    if [ $1 -le 9 ]; then echo "00000$1"
    elif  [ $1 -le 99 ]; then echo "0000$1"
    elif  [ $1 -le 999 ]; then echo "000$1"
    elif  [ $1 -le 9999 ]; then echo "00$1"
    elif  [ $1 -le 99999 ]; then echo "0$1"
    elif  [ $1 -le 999999 ]; then echo "$1"
    fi	
}

# Return the given number of blocks of the given cdrom device or iso file to
# stdout.
# Uses dd to read, because the Linux kernel drivers for cdroms do not recognise
# EOF for something like "cat /dev/cdrom" correctly (some kernels crash(!), some
# return too much data).
# iso_read_blocks DEVICE_OR_ISOFILE NUMBLOCKS
iso_read_blocks() {
    dd if="$1" bs=2048 count="$2"
    # block size does not affect speed:
    #> sh -c 'time dd if=/dev/hdc of=/dev/null bs=2048 count=25000'
    #25000+0 records in
    #25000+0 records out
    #real    1m3.684s
    #user    0m0.060s
    #sys     0m1.000s
    #> sh -c 'time dd if=/dev/hdc of=/dev/null bs=20480 count=2500'
    #2500+0 records in
    #2500+0 records out
    #real    1m3.627s
    #user    0m0.010s
    #sys     0m0.610s
    #> sh -c 'time dd if=/dev/hdc of=/dev/null bs=10k count=5000'
    #real    1m2.415s
}

# Establish the device of a CD mounted at DIR. DIR can be a symlink.
# get_cd_device_from_mountpoint DIR
# Out: CD_MOUNTPOINT	the actual mount point of the CD
#      CD_DEVICE    	the corresponding device
get_cd_device_from_mountpoint() {
    CD_MOUNTPOINT="" CD_DEVICE=""
    test -d "$1/." || return
    pushd "$1/." >/dev/null || return	# can't cd to mountpoint
    CD_MOUNTPOINT="`pwd -P`"
    popd >/dev/null
    find_better_prog
    CD_DEVICE="$(mount | AWK '$3 ~ "'"$CD_MOUNTPOINT"'" {print $1}')"
}

# Read given cdrom device or iso file and write corresponding md5 checksum to
# stdout. The filename with the checksum is always "-".
iso_create_md5() {
    iso_blockread "$1" | md5sum -
}

# Check the given cdrom device or iso file with the given md5 checksum.
# The checksum may be read from stdin (use "-" as filename), but then a temp
# file is required. The temp file may be given as $3 in this case.
# The filename with the md5 checksum must always be "-".
# iso_check_md5 DEVICE_OR_ISOFILE MD5_FILE [TMPFILE_FOR_MD5]
iso_check_md5() {
    if [ "x$2" == "x-" ]; then
    	local tmpfile="${3-$HOME/.functions_vk.bash.iso_check_md5.$$}" stat
	cat - >"$tmpfile"
	iso_blockread "$1" | md5sum -c "$tmpfile"
	stat=$?
	rm "$tmpfile"
	return $stat
    else
    	iso_blockread "$1" | md5sum -c "$2"
	return $?
    fi
}

# Functions to load and eject the CDROM tray
# PLEASE do not edit this file, redefine these functions after loading this.
cdrom_tray_open() {
    type -p eject >/dev/null && eject
}
cdrom_tray_close() {
    type -p eject >/dev/null && eject -t
}


#### Etc ####

# Filter; gunzips hex code to stdout; hex code can be inlined
# Create hex code with one of:
#   gzip -9 | od -w39 -A o -t x1 | sed -e 's/[0-7]*//' -e 's/ //g' -e '/^$/d'
#   gzip -9 | bin2hex | cut -c -50 | sed -e 's, ,,g'
# or simply call createcontainer with desired hex-width as optional argument.
# Call like this:
#   writecontainer <<EOF
#   <hex code>
#   EOF
createcontainer() {
    gzip -9 | od -w${1-39} -A o -t x1 \
    	| sed -e 's/[0-7]*//' -e 's/ //g' -e '/^$/d'
}
writecontainer() {
    #sed -e 's,..,\\\\x&,g' | while read; do
    #	echo 'echo -ne "'"$REPLY"\"
    #done | $BASH -s | gunzip
    sed -e 's,..,\\x&,g' -e 's,^.*$,echo -ne "&",' | $BASH -s | gunzip
}

# display date + time, GMT + local
# Use   Date() { Date1; }   to default to Date1 for Date.
Date() { Date2; }
Date2() {
    date -u +'%a %d %b %Y %T %Z'
    date    +'%a %d %b %Y %T %Z'
}
Date1() {
    echo "`date -u +'%a %d %b %Y %T %Z'`" "   `date +'%a %d %b %Y %T %Z'`"
}

# Display time at start and finish.
#   show_start_time
#   save=$_save_seconds
#   ...
#   show_finish_time $save
# Or, if certain that there is no nested calling:
#   show_start_time
#   ...
#   show_finish_time
# Never write $SECONDS!!!
show_start_time() {
    Date
    _save_seconds=$SECONDS
}
show_finish_time() {
    local sec=$SECONDS
    sec=$(($sec - ${1-$_save_seconds}))
    Date
    echo "Total execution time: `sec2h $sec`"
}

# convert seconds to HH:MM:SS
sec2h() {
    printf "%2i:%02i:%02i" $(($1 / 3600)) $((($1 / 60) % 60)) $(($1 % 60))
}

# ringbell [N] [TEXT]
# Ring terminal bell N times (default 1),
# displaying TEXT (default "Ring bell:\n").
ringbell() {
    local rings="${1:-1}" bell="\007${2-Ring bell:\n}"
    while [ $rings -gt 0 ]; do
    	echo -e -n "$bell"
	test $rings -gt 1 && sleep 1
	rings=$(($rings - 1))
    done
}

# Print program version $VERSION
version() {
    echo "${0##*/} version $VERSION"
}

# Check given number (e.g. previous exit status) and exit with that number if
# not 0
CheckStatus() {
    test "$1" -ne 0 && exit "$1"
}

# cd to dir, if given; then print current working dir in any case
Pwd() {
    test -n "$1" && {
        cd "$1" || exitwith ErrNoDir "$1"
    }
    echo "Current dir is now:" `pwd`
}

# printf %02d
digit02() {
    test ${#1} = 1 && echo "0$1" || echo "$1"
}

# Character constants
HT="	" NL="
" CR="
"
eval `echo -e -n BEL="\007" BS="\010" NP="\014" ESC="\033"`

# Assemble words into an alternate regular expression
assemble_alternate() {
    local re="$1" mid="$2" end="$3" first=1
    shift 3
    while [ $# -gt 0 ]; do
    	test -n "$first" && { re="$re$1"; first=; } || re="$re$mid$1"
	shift
    done
    REGEX="$re$end"
}
assemble_regex() {
    assemble_alternate "(" "|" ")" "$@"
}


#### Debug the lot ####

debug_functions_vk() {
    find_GNU_prog
    find_better_prog
    readresources FILE
    debug_find_GNU_prog
    debug_ifset
    debug_show_args -d 2 "3 3" 4 "" 6 "'"
    path_debug
    debug_shopt_stack
}
